import pandas as pd
from scipy.stats import pearsonr
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, RBF, WhiteKernel
from sklearn.metrics import make_scorer, mean_absolute_error, mean_squared_error
from sklearn.model_selection import train_test_split, GridSearchCV
import numpy as np
import pickle
import torch
from transformers import AlbertTokenizer, AlbertModel, AlbertConfig
from argparse import ArgumentParser
import gensim

np.random.seed(0)

model_path = "albert-base-v2"
do_lower_case = True
config = AlbertConfig.from_pretrained(model_path, output_hidden_states=True)
tokenizer = AlbertTokenizer.from_pretrained(model_path, do_lower_case=do_lower_case)
model = AlbertModel.from_pretrained(model_path, config=config)

def load_and_structure_dataset(filename: str):
    df = pd.read_csv(filename)
    return df['STATUS'].values.tolist(), df['sEXT'].values.tolist(), df['sAGR'].values.tolist(), df['sNEU'].values.tolist(), df['sOPN'].values.tolist(), df['sCON'].values.tolist()

# Tokenize documents and encode with index tokens
# Input: Array of documents (texts)
def encode_documents(documents):
  encoded_document = [tokenizer.encode(i) for i in documents]
  # Truncate document to input size 512
  for i in range(len(encoded_document)):
    while len(encoded_document[i]) > 512:
        encoded_document[i].pop(-2)
  return encoded_document

# Pad encoded documents to equal size
# Input: Array of encoded documents
# Output: Documents with padding and segment IDs (attention mask)
def pad_documents(encoded_documents):
  document_length = [len(i) for i in encoded_documents]
  max_length = max(document_length)
  pad_id = tokenizer.pad_token_id
  padded_documents = [i + [pad_id] * (max_length - len(i)) for i in encoded_documents]
  segment_ids = [[1] * len(i) + [pad_id] * (max_length - len(i)) for i in encoded_documents]
  return padded_documents, segment_ids

# Generate word embedding for a single encoded document.
# Input: Padded documents with corresponding segment IDs
# Output: Hidden states generated by ALBERT model (word embeddings, size 13 x 1 x length_document x 768 (hidden units/features))
def word_embedding(encoded_document, segment_ids):
    # Put model in evaluation mode
    model.eval()
    with torch.no_grad():
        # Create tensor from tokens and segment_ids
        token_tensor = torch.tensor([encoded_document])
        segment_tensor = torch.tensor([segment_ids])
        
        # Run documents through ALBERT and fetch hidden states
        hidden_states = model(input_ids=token_tensor, attention_mask=segment_tensor)[2]

    return hidden_states

# Generate document embedding by averaging word embeddings generated for document
# Input: Hidden states generated by word embedding function
# Output: Embedded document as average of word embeddings
def document_embedding(token_vectors):
  # `token_vecs` is a tensor with shape [num_tokens x 768]
  token_vecs = token_vectors[-11][0]
  # Calculate the average of all |num_tokens| token vectors.
  sentence_embedding = torch.mean(token_vecs, dim=0)
  return sentence_embedding

# Generate document embeddings for documents in corpus
# Input: Array of documents
# Output: Array with embedded features (X)
def generate_embedding_vectors(documents):
  embedded_features = []
  encoded_documents = encode_documents(documents)
  padded_documents, segment_ids = pad_documents(encoded_documents)

  for i in range(len(padded_documents)):
    embedded_words = word_embedding(padded_documents[i], segment_ids[i])
    embedded_document = document_embedding(embedded_words)
    embedded_document = embedded_document.tolist()
    embedded_features.append(embedded_document)
  return embedded_features

def predict(model, X_test, y_test):
    print("Predictions for model ...")
    y_pred = model.predict(X_test)
    print("MSE: " + str(mean_squared_error(y_test, y_pred)))
    print("MAE: " + str(mean_absolute_error(y_test, y_pred)))
    print("Pearson: " + str(pearsonr(y_test, y_pred)[0]))

def custom_scorer(model, X, y):
  y_pred = model.predict(X)
  y_true = y
  return pearsonr(y_true, y_pred)[0]
  
grid_scorer = make_scorer(custom_scorer, greater_is_better=True)

def train_model(X_train, y_train, param):
    clf_gp = GaussianProcessRegressor()
    grid = GridSearchCV(clf_gp, param, refit=True, verbose=3, scoring=custom_scorer)
    grid.fit(X_train, y_train)
    return grid

def main(args):
    use_idun = args.idun
    save_model = args.savemodel
    train_size = 0.9

    param_grid = {
    'kernel': [DotProduct() + WhiteKernel(), RBF(length_scale=1.0) + WhiteKernel()],
    }

    data_path = '/cluster/home/larsmagu/masteroppgave/data/dataset/personality_d.csv' if use_idun else 'personality_d.csv'

    X, y_ext, y_agr, y_neu, y_opn, y_con = load_and_structure_dataset(data_path)
    X = [gensim.parsing.preprocessing.remove_stopwords(text) for text in X]
    extracted_features = generate_embedding_vectors(X)

    X_train_ext, X_test_ext, y_train_ext, y_test_ext = train_test_split(extracted_features, y_ext, train_size = train_size, random_state=42)
    X_train_agr, X_test_agr, y_train_agr, y_test_agr = train_test_split(extracted_features, y_agr, train_size = train_size, random_state=42)
    X_train_neu, X_test_neu, y_train_neu, y_test_neu = train_test_split(extracted_features, y_neu, train_size = train_size, random_state=42)
    X_train_opn, X_test_opn, y_train_opn, y_test_opn = train_test_split(extracted_features, y_opn, train_size = train_size, random_state=42)
    X_train_con, X_test_con, y_train_con, y_test_con = train_test_split(extracted_features, y_con, train_size = train_size, random_state=42)

    model_path = '/cluster/home/larsmagu/masteroppgave/model_training/trained_models/gp_albert'

    print("Training EXTRAVERSION....")
    model_ext = train_model(X_train_ext, y_train_ext, param_grid)
    print("Training AGREEABLENESS...")
    model_agr = train_model(X_train_agr, y_train_agr, param_grid)
    print("Training NEUROTICISM...")
    model_neu = train_model(X_train_neu, y_train_neu, param_grid)
    print("Training OPENNESS...")
    model_opn = train_model(X_train_opn, y_train_opn, param_grid)
    print("Training CONSCIENTIOUSNESS")
    model_con = train_model(X_train_con, y_train_con, param_grid)

    if save_model and use_idun:
        with open(model_path + '/model_ext_prediction2_layer2.pkl', 'wb') as f:
            pickle.dump(model_ext, f)
        with open(model_path + '/model_agr_prediction2_layer2.pkl', 'wb') as f:
            pickle.dump(model_agr, f)
        with open(model_path + '/model_neu_prediction2_layer2.pkl', 'wb') as f:
            pickle.dump(model_neu, f)
        with open(model_path + '/model_opn_prediction2_layer2.pkl', 'wb') as f:
            pickle.dump(model_opn, f)
        with open(model_path + '/model_con_prediction2_layer2.pkl', 'wb') as f:
            pickle.dump(model_con, f)

    # Print prediction for models
    print("------=PREDICTIONS FOR MODELS=------")
    predict(model_ext, X_test_ext, y_test_ext)
    predict(model_agr, X_test_agr, y_test_agr)
    predict(model_neu, X_test_neu, y_test_neu)
    predict(model_opn, X_test_opn, y_test_opn)
    predict(model_con, X_test_con, y_test_con)

    print("------=BEST PARAMETERS FOR MODELS=------")
    print("Parameters EXTRAVERSION: " + str(model_ext.best_params_))
    print("Parameters AGREEABLENESS: " + str(model_agr.best_params_))
    print("Parameters NEUROTICISM: " + str(model_neu.best_params_))
    print("Parameters OPENNESS: " + str(model_opn.best_params_))
    print("Parameters CONSCIENTIOUSNESS: " + str(model_con.best_params_))

if __name__ == "__main__":
    print('Starting ...')

    parser = ArgumentParser()
    parser.add_argument("--idun", default=False, type=bool)
    parser.add_argument("--savemodel", default=True, type=bool)
    args = parser.parse_args()
    main(args)
    print('Complete')

















